/* $Id: module.c,v 1.4 2007/08/28 02:40:30 mgdepalma Exp $
 *
 * Copyright (C) 2007 Softcraft Impresa, LLC <info@softcraft.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; You may only use version 2 of the License,
 * you have no option to use any other version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "module.h"
#include "language.h"

extern gboolean debug;	/* declared in ../programs/<main> */

/*
 * module_load
 */
Modulus *
module_load (const gchar *file, guint spaces, gpointer data)
{
  Modulus *applet = NULL;
  GModule *module = g_module_open (file, G_MODULE_BIND_MASK);

  if (module) {
    gpointer module_init, module_open, module_close;
    applet = g_new0 (Modulus, 1);

    /* Mandatory symbols of a struct _modulus */
    if (!g_module_symbol(module, "module_init", &module_init)) {
      g_warning("Error loading module: %s\n", file);
      g_return_val_if_reached(NULL);
    }

    applet->module_init = module_init;
    applet->module_init (applet, data);	/* initialize plugin module */

    if (applet->space & spaces) {	/* must belong in the spaces mask */
      if (g_module_symbol(module, "module_open", &module_open))
        applet->module_open = module_open;

      if (g_module_symbol(module, "module_close", &module_close))
        applet->module_close = module_close;

      applet->module = module;		/* save struct _GModule */
    }
    else {
      g_free (applet);
      applet = NULL;
    }
  }
  else {
    g_warning(g_module_error());
    g_return_val_if_reached(NULL);
  }
  return applet;
} /* module_load */

/*
 * module_search
 */
Modulus *
module_search (GList *moduli, const gchar* name)
{
  GList *iter;
  Modulus *module = NULL;

  for (iter = moduli; iter != NULL; iter = iter->next) {
    if (strcmp(((Modulus *)iter->data)->name, name) == 0) {
      module = (Modulus *)iter->data;
      break;
    }
  }
  return module;
} /* </module_search> */

/*
 * moduli_remove
 */
GList *
moduli_remove (GList *moduli, Modulus *applet)
{
  GList *list = moduli;

  g_list_remove (list, applet);
  g_free (applet);

  return list;
} /* </moduli_remove> */

/*
 * moduli_space loads all modules from given path (and space(s))
 */
GList *
moduli_space (GList *moduli, const gchar *path, guint spaces, gpointer data)
{
  struct stat current;
  struct stat loaded;

  if (lstat(path, &current) == 0) {	/* we must be able to stat path */
    struct dirent **names;
    Modulus *applet, *modulus;
    int count = scandir(path, &names, NULL, alphasort);
    char file[FILENAME_MAX];
    char *name, *scan;
    int idx;

    for (idx = 0; idx < count; idx++) {
      name = names[idx]->d_name;

      if (name[0] == '.' || strcmp(strrchr(name, '.'), ".so"))
        continue;
      
      strcpy(file, name);	/* isolate file name without extension */
      scan = strrchr(file, '.');
      *scan = (char)0;

      /*
       * See if a module with the same name is already loaded.
       */
      if ((applet = module_search(moduli, file)) != NULL) {
        sprintf(file, "%s/%s", path, name);

        if (lstat(file, &current) != 0)	/* should not happen */
          continue;

        if (lstat(g_module_name(applet->module), &loaded) != 0)
          continue;

        if (loaded.st_mtime > current.st_mtime) {
          continue;	/* previously loaded module more current */
        }
        else {
          if ((modulus = module_load(file, spaces, data)) != NULL)
            moduli_remove(moduli, applet);
          else
            continue;	/* stick with previously loaded module */
        }
      }
      else {
        sprintf(file, "%s/%s", path, name);
        modulus = module_load(file, spaces, data);
      }

      if (modulus) { /* NULL if not found or not in spaces mask */
        moduli = g_list_append (moduli, modulus);
      }
    }
  }
  return moduli;
} /* </moduli_space> */
