/* $Id: window.c,v 1.11 2007/09/07 21:38:22 mgdepalma Exp $
 *
 * Copyright (C) 2007 Softcraft Impresa, LLC <info@softcraft.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; You may only use version 2 of the License,
 * you have no option to use any other version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "window.h"
#include "grabber.h"
#include "xpmglyphs.h"
#include "language.h"

extern gboolean debug;	/* declared elsewhere */


/*
 * file_path_find - search for file name from a path list
 */
const gchar *
file_path_find (GList *list, const gchar *name)
{
  static gchar *file[FILENAME_MAX];
  gchar *pathname = NULL;
  GList *iter;

  for (iter = list; iter != NULL; iter = iter->next) {
    sprintf((gchar *)file, "%s/%s", (gchar *)iter->data, (gchar *)name);

    if (g_file_test((gchar *)file, G_FILE_TEST_EXISTS)) {
      pathname = (gchar *)file;
      break;
    }
  }
  return pathname;
} /* </file_path_find> */

/*
 * create_mask_from_pixmap - create the bitmap mask for a pixmap.
 */
GdkBitmap *
create_mask_from_pixmap (GdkPixmap *pixmap, gint width, gint height)
{
  GdkBitmap *mask = NULL;
  GdkImage *image = gdk_drawable_get_image (pixmap, 0, 0, width, height);
  guint32 pixelmask = gdk_image_get_pixel (image, 0, 0);

  /* Bitmap mask, one bit per pixel, each row starts on a byte boundary. */
  size_t length = (width + 7) / 8 * height;  /* bits per line * height */
  guint8 *data = g_new (guint8, length);
  memset (data, 255, length);	/* set bits are unmasked */
  int row, col;
  int idx = 0;

  for (row = 0; row < height; row++) {
    for (col = 0; col < width; col++, idx++) {
      //if (debug) printf("%x ", gdk_image_get_pixel (image, col, row));
      if (gdk_image_get_pixel (image, col, row) == pixelmask)
        data[idx >> 3] ^= 1 << (idx & 7);
    }
    //if (debug) printf("\n");
    idx = (idx + 7) & ~7u;	/* move index to next byte boundary */
  }

  mask = gdk_bitmap_create_from_data (NULL, (gchar*)data, width, height);
  g_free (data);

  return mask;
} /* </create_mask_from_pixmap> */

/*
 * get_label_size
 */
GtkAllocation *
get_label_size (GtkWidget *label)
{
  gint xsize, ysize;
  GtkAllocation *allocation = g_new0(GtkAllocation, 1);

  PangoLayout *layout = gtk_label_get_layout (GTK_LABEL(label));
  pango_layout_get_pixel_size (layout, &xsize, &ysize);
  allocation->height = ysize;
  allocation->width = xsize;

  return allocation;
} /* </get_label_size> */

/*
 * image_new_from_file_scaled - returns a scaled image given a file name
 */
GtkWidget *
image_new_from_file_scaled (const gchar *file, gint width, gint height)
{
  GtkWidget *image = NULL;
 
  if (file != NULL && g_file_test(file, G_FILE_TEST_EXISTS)) {
    GdkPixbuf *pixbuf = pixbuf_new_from_file_scaled (file, width, height);

    if (pixbuf) {
      image = gtk_image_new_from_pixbuf (pixbuf);
      g_object_unref (pixbuf);
    }
  }

  if (image == NULL)
    image = gtk_image_new_from_stock(GTK_STOCK_MISSING_IMAGE,
                                        GTK_ICON_SIZE_BUTTON);
  return image;
} /* </image_new_from_file_scaled> */

/*
 * pixbuf_new_from_file_scaled - return scaled GdkPixbuf from a file
 */
GdkPixbuf *
pixbuf_new_from_file_scaled (const gchar *file, gint width, gint height)
{
  GError    *error  = NULL;
  GdkPixbuf *image  = gdk_pixbuf_new_from_file(file, &error);
  GdkPixbuf *pixbuf = NULL;

  if (image) {
    pixbuf = pixbuf_scale (image, width, height);
    g_object_unref (image);
  }
  return pixbuf;
} /* </pixbuf_new_from_file_scaled> */

/*
 * pixbuf_scale - scale a GdkPixbuf to given width and height
 */
GdkPixbuf *
pixbuf_scale (GdkPixbuf *pixbuf, gint width, gint height)
{
  return gdk_pixbuf_scale_simple(pixbuf, width, height, GDK_INTERP_BILINEAR);
} /* </pixbuf_scale> */

/*
 * redraw_pixbuf - redraw GdkPixbuf of a canvas drawing area
 */
gboolean
redraw_pixbuf (GtkWidget *canvas, GdkPixbuf *image)
{
  gint xsize, ysize;
  gboolean vote = FALSE;

  g_return_val_if_fail (GDK_IS_DRAWABLE(canvas->window), FALSE);

  /* Obtain width and height of the canvas area. */
  gdk_drawable_get_size (canvas->window, &xsize, &ysize);
  gdk_window_clear_area (canvas->window, 0, 0, xsize, ysize);

  if (image) {
    GdkPixbuf *pixbuf;
    gboolean   scaled;	/* we may create a new scaled pixbuf */

    gint width  = gdk_pixbuf_get_width (image);
    gint height = gdk_pixbuf_get_height (image);
    gint xorg, yorg;                            /* start (x,y) coordinates */

    float factor = fmax((float)width / xsize, (float)height / ysize) * 0.99;

    /* Adjust width and height by scale factor. */
    if (width > xsize || height > ysize) {
      width  = (float)width  / factor;
      height = (float)height / factor;
      pixbuf = pixbuf_scale (image, width, height);
      scaled = TRUE;
    }
    else {
      pixbuf = image;
      scaled = FALSE;
    }

    /* Adjust image (x,y) start coordinates. */
    xorg = abs(xsize - width) / 2;
    yorg = abs(ysize - height) / 2;

    /* Draw scaled pixbuf onto the canvas area. */
    gdk_draw_pixbuf (canvas->window,
                     canvas->style->fg_gc[GTK_WIDGET_STATE (canvas)],
                     pixbuf, 0, 0, xorg, yorg, width, height,
                     GDK_RGB_DITHER_NONE, 0, 0);
    if (scaled)
      g_object_unref (pixbuf);

    vote = TRUE;
  }
  return vote;
} /* </redraw_pixbuf> */

/*
 * set_background_pixbuf - set background pixmap of window using pixbuf
 */
void
set_background_pixbuf (GdkWindow *window, GdkPixbuf *pixbuf)
{
  GdkGC     *gc;
  GdkPixbuf *image;
  GdkPixmap *pixmap;
  gint width, height;
  gint xres, yres;


  /* Get the window and image sizes */
  gdk_window_get_size (window, &xres, &yres);
  height = gdk_pixbuf_get_height (pixbuf);
  width  = gdk_pixbuf_get_width (pixbuf);

  /* Scale GdkPixbuf *image as needed */
  if (width != xres || height != yres)
    image = pixbuf_scale (pixbuf, xres, yres);
  else
    image = pixbuf;

  /* Create pixmap at window size and render image */
  gc = gdk_gc_new (window);
  pixmap = gdk_pixmap_new (window, xres, yres, gdk_visual_get_best_depth());
  gdk_pixbuf_render_to_drawable (image, pixmap, gc, 0, 0, 0, 0, xres, yres,
                                 GDK_RGB_DITHER_NORMAL, 1, 1);
  g_object_unref (gc);

  /* Set background pixmap on window */
  gdk_window_set_back_pixmap (window, pixmap, FALSE);
  g_object_unref (pixmap);

  gdk_window_clear(window);   /* clear window to effect change */
  gdk_flush();
} /* </set_background_pixbuf> */

/*
 * draw_pixbuf
 */
void
draw_pixbuf (GdkDrawable *target,
             GdkPixbuf *pixbuf, GdkGC *gc,
             gint xpos, gint ypos)
{
  gdk_draw_pixbuf (target,	/* destination drawable */
                   gc,		/* GdkGC, used for clipping, or NULL */
                   pixbuf,	/* source GdkPixbuf */
                   0, 0,	/* source (x,y) coordinates */
                   xpos, ypos,	/* destination (x,y) coordinates */
                   -1, -1,	/* dimensions or -1 to use pixbuf size */
                   GDK_RGB_DITHER_NORMAL,
                   0, 0);	/* dither (x,y) coordinates */
} /* </draw_pixbuf> */

/*
 * draw_pixmap - set/unset bits
 *
 * sample usage:
 *
  gdk_pixbuf_render_pixmap_and_mask (pixbuf, &pixmap, &mask, 255);
  GdkBitmap *inverse = bitmap_new_with_pixel (window, width, height, 1);
  GdkGC *gc = gdk_gc_new (inverse);

  // Tells the Graphics context to draw using GDK_CLEAR.
  // That is, everywhere it wishes to draw it will actually
  // clear the associated pixmap (in this case bitmap) instead
  gdk_gc_set_function(gc, GDK_CLEAR);

  // Tells the Graphics Context to ONLY draw the places where
  // mask has been drawn on.
  gdk_gc_set_clip_mask(gc, mask);

  // Make the call.
  draw_pixmap (inverse, mask, gc, 0, 0);
 */
void
draw_pixmap (GdkDrawable *target,
             GdkPixmap *pixmap, GdkGC *gc,
             gint xpos, gint ypos)
{
  gdk_draw_drawable(
    target,	/* GdkDrawable *destination */
    gc,		/* GdkGC, used for clipping, or NULL */
    pixmap,	/* GdkDrawable *source */
    0, 0,	/* source (x,y) coordinates */
    xpos, ypos,	/* destination (x,y) coordinates */
    -1, -1	/* width and height; (-1,-1) => source dimensions */
  );
} /* </draw_pixmap> */

/*
 * sticky_window_new - create a sticky GTK_WINDOW_TOPLEVEL window
 */
GtkWidget *
sticky_window_new (GdkWindowTypeHint hint,
                   gint width, gint height,
                   gint xpos,  gint ypos)
{ 
  GtkWidget *widget;
  GtkWindow *window;

  /* Create a new window and set attributes. */
  widget = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  window = GTK_WINDOW (widget);

  gtk_window_set_default_size (window, width, height);
  gtk_window_set_decorated (window, FALSE);
  gtk_window_set_resizable (window, FALSE);

  gtk_window_set_skip_pager_hint (window, TRUE);
  gtk_window_set_skip_taskbar_hint (window, TRUE);
  gtk_window_set_type_hint (window, hint);

  gtk_window_move (window, xpos, ypos);
  gtk_window_stick(window);

  if (hint == GDK_WINDOW_TYPE_HINT_DOCK || hint == GDK_WINDOW_TYPE_HINT_TOOLBAR)
  {
    /* http://standards.freedesktop.org/wm-spec/1.3/ar01s05.html */
    GdkAtom card  = gdk_atom_intern("CARDINAL",      FALSE);
    GdkAtom strut = gdk_atom_intern("_NET_WM_STRUT", FALSE);

    long wm_strut[] = { 0,			/* Left   */
                        0,			/* Right  */
                        height,			/* Top    */
                        0,			/* Bottom */
                      };

    gdk_property_change (GDK_WINDOW (widget->window), strut, card,
                         32, GDK_PROP_MODE_REPLACE,
                         (guchar *)(&wm_strut), 4);

    /* Explicitly keep the window above all others. */
    gtk_window_set_keep_above (window, TRUE);
  }
  gtk_widget_set_size_request (widget, width, height);

  /* Set event handlers for window close/destroy */
  g_signal_connect (G_OBJECT (widget), "destroy",
                      G_CALLBACK (gtk_widget_destroy), NULL);

  g_signal_connect (G_OBJECT (widget), "delete_event",
                      G_CALLBACK (gtk_widget_destroy), NULL);
  return widget;
} /* </sticky_window_new> */
