/* $Id: xmlconfig.c,v 1.9 2007/09/30 15:41:10 mgdepalma Exp $
 *
 * Copyright (C) 2007 Softcraft Impresa, LLC <info@softcraft.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; You may only use version 2 of the License,
 * you have no option to use any other version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "xmlconfig.h"


/*
 * configuration_attrib
 * configuration_attrib_string
 */
gchar *
configuration_attrib (ConfigurationNode *node, gchar *name)
{
  ConfigurationAttrib *attrib = node->attrib;
  gchar *value = NULL;

  while (attrib != NULL) {
    if (strcmp(attrib->name, name) == 0) {
      value = attrib->value;
      break;
    }
    attrib = attrib->next;
  }
  return value;
} /* </configuration_attrib> */

gchar *
configuration_attrib_string (ConfigurationNode *node)
{
  ConfigurationAttrib *attrib = node->attrib;
  static gchar line[FILENAME_MAX];
  gchar *string = NULL;

  if (attrib != NULL) {
    strcpy(line, attrib->name);
    strcat(line, "=\"");
    strcat(line, attrib->value);
    strcat(line, "\"");

    for (attrib = attrib->next; attrib != NULL; attrib = attrib->next) {
      strcat(line, " ");
      strcat(line, attrib->name);
      strcat(line, "=\"");
      strcat(line, attrib->value);
      strcat(line, "\"");
    }
    string = line;
  }
  return string;
} /* </configuration_attrib_string> */

/*
 * configuration_find
 * configuration_find_next
 */
ConfigurationNode *
configuration_find (ConfigurationNode *config, const gchar *element)
{
  ConfigurationNode *item;

  for (item = config; item != NULL; item = item->next) {
    if (item->depth < config->depth || item->element == NULL) {
      item = NULL;
      break;
    }
    if (strcmp(item->element, element) == 0) {
      if (item->next && (item->next)->type == XML_READER_TYPE_ATTRIBUTE)
        item = item->next;
      break;
    }
  }
  return item;
} /* </configuration_find> */

ConfigurationNode *
configuration_find_next (ConfigurationNode *config, const gchar *element)
{
  ConfigurationNode *chain = configuration_find (config, element);
  ConfigurationNode *node  = NULL;

  if (chain != NULL) {
    guint depth = chain->depth;

    while (chain != NULL) {
      if (chain->type == XML_READER_TYPE_END_ELEMENT && chain->depth == depth) {
        node = chain->next;
        break;
      }
      chain = chain->next;
    }
  }
  return node;
} /* </configuration_find_next> */

/*
 * configuration_insert
 * configuration_remove
 */
gboolean
configuration_insert (ConfigurationNode *config,
                      ConfigurationNode *node,
                      gboolean nest)
{
  ConfigurationNode *iter;
  ConfigurationNode *chain;
  ConfigurationNode *trail;

  for (iter = node; iter != NULL; iter = iter->next) {
    if (nest) { 		/* correct nesting depth */
      iter->depth += config->depth;
    }

    if (iter->next == NULL) {	/* save the last node on the chain */
      trail = iter;
      break;
    }
  }

  chain = config->next;	  	/* save forward link */
  chain->back  = trail;
  config->next = node;

  node->back = config;
  trail->next = chain;

  return TRUE;
} /* </configuration_insert> */

gboolean
configuration_remove (ConfigurationNode *node)
{
  ConfigurationNode *mark;
  gchar tail[FILENAME_MAX];
  gboolean vote = FALSE;

  sprintf(tail, "/%s", node->element);
  mark = configuration_find (node, tail);  /* corresponding end element */

  if (mark != NULL) {
    ConfigurationNode *item;
    ConfigurationNode *chain;
    ConfigurationNode *trail;
    GList *list = NULL;
    GList *iter;

    /* Prune and graft to previous node and after the end element */
    chain = node->back;
    trail = mark->next;

    chain->next = trail;
    trail->back = chain;

    for (item = node; item != mark->next; item = item->next)
      list = g_list_append (list, item);

    for (iter = list; iter != NULL; iter = iter->next) {
      item = (ConfigurationNode *)iter->data;

      g_free (item->element);
      if (item->attrib) g_free (item->attrib);
      g_free (item);
    }

    g_list_free (list);
    vote = TRUE;
  }
  return vote;
} /* </configuration_remove> */

/*
 * configuration_path
 */
gchar *
configuration_path (ConfigurationNode *config)
{
  ConfigurationNode *item = configuration_find (config, "path");
  gchar *path = NULL;

  if (item != NULL) {
    if ((item->element)[0] == '$') {	/* getenv (&(item->element)[1]) */
      char dirname[FILENAME_MAX];
      char *scan;

      strcpy(dirname, item->element);
      scan = strchr(dirname, '/');

      if (scan) {
        *scan = (char)0;
        path = g_strdup_printf("%s/%s", getenv(&dirname[1]), scan+1);
      }
      else {
        path = getenv(&dirname[1]);
      }
    }
    else {
      path = item->element;
    }
  }
  return path;
} /* </configuration_path> */

/*
 * (private) configuration_read_attributes
 */
static ConfigurationAttrib *
configuration_read_attributes (xmlTextReaderPtr reader)
{
  int count = xmlTextReaderAttributeCount(reader);
  ConfigurationAttrib *attrib = g_new0(ConfigurationAttrib, 1);
  ConfigurationAttrib *chain  = attrib;
  int idx;

  for (idx = 0; idx < count; idx++) {
    xmlTextReaderMoveToAttributeNo(reader, idx);

    chain->name  = g_strdup ((gchar *)xmlTextReaderConstName(reader));
    chain->value = g_strdup ((gchar *)xmlTextReaderConstValue(reader));

    if (idx < count - 1)
      chain = chain->next = g_new0(ConfigurationAttrib, 1);
  }
  xmlTextReaderMoveToElement(reader);	/* need to move back to the element */

  return attrib;
} /* </configuration_read_attributes> */

/*
 * configuration_read
 */
ConfigurationNode *
configuration_read (const gchar *data, const gchar *schema, gboolean memory)
{
  ConfigurationNode *config = NULL;
  xmlTextReaderPtr reader;

  if (memory)
    reader = xmlReaderForMemory (data, strlen(data), NULL, schema, 0);
  else
    reader = xmlReaderForFile(data, NULL, 0);

  if (reader) {
    ConfigurationNode *item, *node;

    int status = xmlTextReaderRead(reader);
    const xmlChar *name, *value;

    if (memory == FALSE)
      if (strcmp((char *)xmlTextReaderConstName(reader), schema)) {
        fprintf(stderr, "%s: invalid configuration file.\n", data);
        return NULL;
      }

    /* Start a new ConfigurationNode data structure. */
    config = g_new0(ConfigurationNode, 1);
    item = config;

    while (status == XML_TEXTREADER_MODE_INTERACTIVE) {
      name  = xmlTextReaderConstName(reader);
      value = xmlTextReaderConstValue(reader);

      item->depth = xmlTextReaderDepth(reader);
      item->type = xmlTextReaderNodeType(reader);

      /* isolate name and value from text string */
      if (xmlTextReaderHasValue(reader)) {
        if (value[0] != '\n') {
          item->type = XML_READER_TYPE_ATTRIBUTE;
          item->element = g_strdup((gchar *)value);
        }
      }
      else {
        if (item->type == XML_READER_TYPE_END_ELEMENT)
          item->element = g_strdup_printf("/%s", (gchar *)name);
        else
          item->element = g_strdup((gchar *)name);
      }

      if (xmlTextReaderHasAttributes(reader)) {
        if (item->type != XML_READER_TYPE_END_ELEMENT)
          item->attrib = configuration_read_attributes (reader);
      }

      /* allocate a ConfigurationNode for the next record */
      if (item->element) {
        node = item;
        item->next = g_new0(ConfigurationNode, 1);
        item = item->next;
        item->back = node;
      }
      status = xmlTextReaderRead(reader);
    }

    /* Last node in the chain is redundant, and could cause errors. */
    if (node->element) {
      node->next = NULL;
      g_free (item);
    }
    xmlFreeTextReader(reader);
  }
  return config;
} /* </configuration_read> */

/*
 * configuration_write
 */
void
configuration_write (ConfigurationNode *config,
                     const char *header,
                     FILE *stream)
{
  const char *padding = "                  ";
  char indent[strlen(padding)];
  ConfigurationNode *item = config;
  gchar *attrib;

  if (header == NULL) {
    fprintf(stream, "<?xml version=\"1.0\" encoding=\"utf-16\"?>\n");
  }
  else {
    fprintf(stream, "%s", header);
    item = item->next->next;   /* strange, comment would be immediately next */
  }

  for (; item != NULL; item = item->next) {
    attrib = configuration_attrib_string (item);
    strncpy(indent, padding, item->depth);
    indent[item->depth] = (char)0;

    if (item->next && (item->next)->type == XML_READER_TYPE_ATTRIBUTE) {
      if (attrib != NULL)
        fprintf(stream, "%s<%s %s>%s</%s>\n", indent, item->element,
                       attrib, (item->next)->element, item->element);
      else
        fprintf(stream, "%s<%s>%s</%s>\n", indent, item->element,
                              (item->next)->element, item->element);
      item = (item->next)->next;
    }
    else if (item->element != NULL) {
      if (item->type == XML_READER_TYPE_ELEMENT &&
          (strcmp(item->element, "menu") == 0 ||
           strcmp(item->element, "modules") == 0 ||
           strcmp(item->element, "desktop") == 0)) fprintf(stream, "\n");

      if (attrib != NULL)
        fprintf(stream, "%s<%s %s>\n", indent, item->element, attrib);
      else {
        if (item->type == XML_READER_TYPE_ELEMENT &&
            strcmp(item->element, "separator") == 0)
          fprintf(stream, "%s<%s />\n", indent, item->element);
        else
          fprintf(stream, "%s<%s>\n", indent, item->element);
      }
    }

    /* Check ahead for the end of configuration. */
    if (header && item->next == NULL)
      break;
  }
} /* </configuration_write */
