/* $Id: battery.c,v 1.6 2007/09/02 23:44:55 mgdepalma Exp $
 *
 * Copyright (C) 2007 Softcraft Impresa, LLC <info@softcraft.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; You may only use version 2 of the License,
 * you have no option to use any other version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "module.h"
#include "gpanel.h"
#include "language.h"

extern gboolean debug;		/* see, gpanel.c */
extern const char *Program;	/* .. */

/*
 * (private) show_battery_state
 * (private) start_battery_monitor
 */
static gboolean
show_battery_state (PanelBattery *battery)
{
  FILE *stream;
  char line[FILENAME_MAX], *state = NULL;
  gboolean discharging = TRUE;
  int remaining;
  int capacity;

  if ((stream = fopen("/proc/acpi/battery/BAT0/state", "r")) != NULL) {
    while (fgets(line, FILENAME_MAX, stream))
      if (strstr(line, "charging state:")) {
        state = &line[strlen("charging state:")];
        while (*state == ' ') ++state;
        state[strlen(state) - 1] = (char)0;
        discharging = (strcmp(state, "discharging") == 0);
      }
      else if (strstr(line, "remaining capacity:")) {
        remaining = atoi(&line[strlen("remaining capacity:")]);
      }

    fclose(stream);
  }

  if ((stream = fopen("/proc/acpi/battery/BAT0/info", "r")) != NULL) {
    while (fgets(line, FILENAME_MAX, stream))
      if (strstr(line, "last full capacity:")) {
        capacity = atoi(&line[strlen("last full capacity:")]);
        break;
      }

    fclose(stream);
  }

  if (capacity <= 0) {	/* safeguard against divide by zero */
    remaining = 100;
    capacity = 100;
  }

  if (state != NULL) {
    guint level = 100 * remaining / capacity;
    gchar *label = g_strdup_printf("%d%c", level, '%');

    if (discharging) {
      redraw_pixbuf(battery->stateImage, battery->dischargingIcon);

      if (level <= battery->critical)
        notice(NULL, ICON_ERROR, "%s: %s %d%c %s.", Program,
               _("battery level critical"), level, '%', _("charge left"));
      else if (level <= battery->warning)
        notice(NULL, ICON_WARNING, "%s: %s %d%c %s.", Program,
               _("battery level low"), level, '%', _("charge left"));
    }
    else
      redraw_pixbuf(battery->stateImage, battery->chargingIcon);

    gtk_label_set_text (GTK_LABEL(battery->stateText), label);
    g_free (label);
  }
  return (battery->interval > 0);
} /* </show_battery_state> */

static gboolean
start_battery_monitor (PanelBattery *battery)
{
  /* Set checking status every 'interval' seconds. */
  g_timeout_add (1000 * battery->interval, show_battery_state, battery);

  return FALSE;
} /* </start_battery_monitor> */

/*
 * Required methods
 */
void
module_init (Modulus *self, GlobalPanel *panel)
{
  self->name    = "battery";
  self->icon    = "battery.png";
  self->release = "1.0";
  self->place   = PACK_END;
  self->space   = MODULI_SPACE_TASKBAR;

  if (access("/proc/acpi/battery/BAT0", F_OK) != 0)
    self->space = 0;	/* disengage applet by brute force */
} /* </module_init> */

/*
 * Optional methods
 */
void
module_open (Modulus *self, GlobalPanel *panel)
{
  GtkWidget *image, *label, *layout;
  PanelBattery *battery = panel->battery;
  PanelIcons *icons = panel->icons;
  guint size = icons->size;
  const gchar *icon;

  icon = file_path_find (icons->path, "power.png");
  battery->chargingIcon = pixbuf_new_from_file_scaled (icon, size, size);

  icon = file_path_find (icons->path, self->icon);
  battery->dischargingIcon = pixbuf_new_from_file_scaled (icon, size, size);

  /* Construct the user interface. */
  layout = gtk_hbox_new (FALSE, 2);

  image = battery->stateImage = gtk_drawing_area_new ();
  gtk_widget_set_size_request (GTK_WIDGET (image), size, size);
  gtk_box_pack_start (GTK_BOX(layout), image, FALSE, FALSE, 0);
  gtk_widget_show (image);

  label = battery->stateText = gtk_label_new ("100%");
  gtk_box_pack_start (GTK_BOX(layout), label, FALSE, FALSE, 0);
  gtk_widget_show (label);

  /* Adjust GlobalPanel pack position. */
  panel_update_pack_position(panel, self->place, label, 0);

  /* Start monitoring battery state. */
  g_timeout_add (1000, start_battery_monitor, battery);

  self->widget = layout;
} /* module_open */

void
module_close (Modulus *self, GlobalPanel *panel)
{
  gtk_widget_destroy (self->widget);
} /* module_close */
