/* $Id: mixer.c,v 1.7 2007/09/10 02:57:16 mgdepalma Exp $
 *
 * Copyright (C) 2007 Softcraft Impresa, LLC <info@softcraft.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; You may only use version 2 of the License,
 * you have no option to use any other version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>

#include "gpanel.h"
#include "module.h"
#include "language.h"

extern gboolean debug;		/* see, gpanel.c */
extern const char *Program;	/* .. */

static GtkWidget *slider;	/* volume control slider window */
static int mixer;		/* file descriptor to /dev/mixer */

/*
 * (private) volume_adjust
 * (private) volume_control
 */
static gboolean
volume_adjust (GtkAdjustment *adjust, GlobalPanel *panel)
{
  gint volume;

  volume  = (gint)adjust->value;
  volume += (gint)adjust->value << 8;

  if ((ioctl(mixer, SOUND_MIXER_WRITE_VOLUME, &volume)) < 0)
    return FALSE;

  return TRUE;
} /* </volume_adjust> */

static void
volume_control (GtkWidget *widget, GlobalPanel *panel)
{
  g_return_if_fail (GTK_IS_WIDGET (widget));

  if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(widget)))
    gtk_widget_show (slider);
  else
    gtk_widget_hide (slider);
} /* </volume_control> */

/*
 * slider_new
 */
static GtkWidget *
slider_new (Modulus *self, GlobalPanel *panel)
{
  GtkAdjustment *adjust;
  GtkWidget *frame, *range, *window;
  guint iconsize = panel->icons->size;
  int vol, xpos, ypos;

  gint width  = gdk_screen_width ();
  gint height = gdk_screen_height ();

  ioctl (mixer, SOUND_MIXER_READ_VOLUME, &vol);

  adjust = (GtkAdjustment*)gtk_adjustment_new (100, 0, 101, 1, 10, 1);
  gtk_adjustment_set_value (adjust, vol >> 8);

  if (panel->orientation == GTK_ORIENTATION_HORIZONTAL) {
    if (self->place == PACK_START)
      xpos = panel->spos + iconsize;
    else
      xpos = panel->epos - iconsize;

    if (panel->place == GTK_POS_BOTTOM)
      ypos = height - panel->height - panel->margin - 104;
    else
      ypos = panel->height + panel->margin;

    range = gtk_vscale_new (adjust);
    gtk_widget_set_usize (GTK_WIDGET(range), iconsize, 100);
  }
  else {
    if (self->place == PACK_START)
      ypos = panel->spos + iconsize;
    else
      ypos = panel->epos - iconsize;

    if (panel->place == GTK_POS_RIGHT)
      xpos = width - panel->width - panel->margin - 100;
    else
      xpos = panel->width + panel->margin;

    range = gtk_hscale_new (adjust);
    gtk_widget_set_usize (GTK_WIDGET(range), 100, iconsize);
  }

  if (panel->place == GTK_POS_BOTTOM || panel->place == GTK_POS_RIGHT)
    gtk_range_set_inverted (GTK_RANGE(range), TRUE);

  gtk_scale_set_digits (GTK_SCALE (range), 0);
  gtk_scale_set_draw_value (GTK_SCALE (range), FALSE);

  /* Create a popup window with a frame. */
  window = gtk_window_new (GTK_WINDOW_POPUP);

  gtk_window_set_keep_above (GTK_WINDOW(window), TRUE);
  gtk_widget_set_uposition (GTK_WIDGET(window), xpos, ypos);

  frame = gtk_frame_new (NULL);
  gtk_frame_set_shadow_type (GTK_FRAME(frame), GTK_SHADOW_OUT);
  gtk_container_add (GTK_CONTAINER(window), frame);
  gtk_widget_show (frame);

  gtk_container_add (GTK_CONTAINER(frame), range);
  g_signal_connect (G_OBJECT(adjust), "value_changed",
                          G_CALLBACK(volume_adjust), adjust);
  gtk_widget_show (range);

  return window;
} /* </slider_new> */

/*
 * Required methods
 */
void
module_init (Modulus *self, GlobalPanel *panel)
{
  self->name    = "mixer";
  self->icon    = "mixer.png";
  self->release = "0.3";
  self->place   = PACK_END;
  self->space   = MODULI_SPACE_TASKBAR;
}  /* </module_init> */

/*
 * module_open - construct user interface
 */
void
module_open (Modulus *self, GlobalPanel *panel)
{
  GtkWidget *image, *layout;
  PanelIcons *icons = panel->icons;
  const gchar *icon;

  if ( (mixer = open("/dev/mixer", O_RDWR)) < 0) {
    panel->notice = g_list_append (panel->notice,
                                   _("Can't open the mixer device."));

    icon = file_path_find (icons->path, "mute.png");
    layout = gtk_button_new();
  }
  else {
    icon = file_path_find (icons->path, self->icon);
    layout = gtk_toggle_button_new();
  }

  /* Construct the user interface. */
  image = image_new_from_file_scaled (icon, icons->size, icons->size);

  gtk_button_set_relief (GTK_BUTTON(layout), GTK_RELIEF_NONE);
  gtk_container_add (GTK_CONTAINER(layout), image);
  gtk_widget_show (image);

  /* Adjust GlobalPanel pack position. */
  panel_update_pack_position(panel, self->place, NULL, icons->size);

  if (mixer > 0) {
    g_signal_connect (G_OBJECT(layout), "toggled",
                            G_CALLBACK(volume_control), panel);

    slider = slider_new (self, panel);  /* volume control slider */
  }
  self->widget = layout;
} /* </module_open> */

void
module_close (Modulus *self, GlobalPanel *panel)
{
  gtk_widget_destroy (self->widget);
  if (mixer) close(mixer);
} /* </module_close> */
